#include <iostream>
#include <string>

// OpenNI lib
#include <XnOpenNI.h>
#include <XnCodecIDs.h>
#include <XnCppWrapper.h>

#include <SDL/SDL.h>

#include "drawer.h"

#define CHECK_RC(nRetVal, what)                 \
  if (nRetVal != XN_STATUS_OK){						\
  std::cerr << what << " failed: " << xnGetStatusString(nRetVal) << std::endl;\
  throw 1; \
  }

#include "event.h"
#define XML_PATH "KinectConfig.xml"

xn::Context g_Context;
xn::UserGenerator g_UserGenerator;


XnBool g_bNeedPose = FALSE;
XnChar g_strPose[20] = "";
XnBool g_bDrawBackground = TRUE;
XnBool g_bDrawPixels = TRUE;
XnBool g_bDrawSkeleton = TRUE;
XnBool g_bPrintID = TRUE;
XnBool g_bPrintState = TRUE;
XnBool g_bPause = false;
XnBool g_bRecord = false;
XnBool g_bQuit = false;

/**
 * Callbacks
 */
// Callback: New user was detected
void XN_CALLBACK_TYPE User_NewUser(xn::UserGenerator& generator, XnUserID nId,
    void* pCookie) {
  #ifdef DEBUG
  std::cout << "New User " << nId << std::endl;
  #endif

  // New user found
  if (g_bNeedPose) { 
    g_UserGenerator.GetPoseDetectionCap().StartPoseDetection(g_strPose, nId);
  } else {
    g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
  }
}

// Callback: An existing user was lost
void XN_CALLBACK_TYPE User_LostUser(xn::UserGenerator& generator, XnUserID nId,
    void* pCookie) {
  #ifdef DEBUG
  std::cout << "Lost user " << nId << std::endl;
  #endif
}

// Callback: Detected a pose
void XN_CALLBACK_TYPE UserPose_PoseDetected(
    xn::PoseDetectionCapability& capability, const XnChar* strPose,
    XnUserID nId, void* pCookie) {
  #ifdef DEBUG
  std::cout << "Pose " << strPose << " detected for user " << nId << std::endl;
  #endif

  g_UserGenerator.GetPoseDetectionCap().StopPoseDetection(nId);
  g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
}

// Callback: Started calibration
void XN_CALLBACK_TYPE UserCalibration_CalibrationStart(
    xn::SkeletonCapability& capability, XnUserID nId, void* pCookie) {
  #ifdef DEBUG
  std::cout << "Calibration started for user " << nId << std::endl;
  #endif
}

// Callback: Finished calibration
void XN_CALLBACK_TYPE UserCalibration_CalibrationEnd(
    xn::SkeletonCapability& capability, XnUserID nId, XnBool bSuccess,
    void* pCookie) {
  if (bSuccess) {
    // Calibration succeeded
    #ifdef DEBUG
    std::cout << "Calibration complete, start tracking user " 
	      << nId << std::endl;
    #endif
    g_UserGenerator.GetSkeletonCap().StartTracking(nId);
  } else {
    // Calibration failed
    #ifdef DEBUG
    std::cout << "Calibration failed for user " << nId << std::endl;
    #endif
    if (g_bNeedPose) {
      g_UserGenerator.GetPoseDetectionCap().StartPoseDetection(g_strPose, nId);
    } else {
      g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
    }
  }
}



Event::Event(std::string xml_conf) { 
  XnStatus nRetVal = XN_STATUS_OK;
  
  xn::EnumerationErrors errors;
  nRetVal = g_Context.InitFromXmlFile(xml_conf.c_str(), &errors);
  
  nRetVal = g_Context.FindExistingNode(XN_NODE_TYPE_USER, g_UserGenerator);
  if (nRetVal != XN_STATUS_OK) {
    nRetVal = g_UserGenerator.Create(g_Context);
    CHECK_RC(nRetVal, "Find user generator");
  }

  XnCallbackHandle hUserCallbacks, hCalibrationCallbacks, hPoseCallbacks;
  if (!g_UserGenerator.IsCapabilitySupported(XN_CAPABILITY_SKELETON)) {
    std::cerr << "Supplied user generator doesn't support skeleton" 
	      << std::endl;
    throw 1;
  }
  
  g_UserGenerator.RegisterUserCallbacks(User_NewUser, 
  					User_LostUser, 
  					NULL,
  					hUserCallbacks);
  g_UserGenerator.GetSkeletonCap().
    RegisterCalibrationCallbacks(UserCalibration_CalibrationStart, 
  				 UserCalibration_CalibrationEnd, 
  				 NULL,				
  				 hCalibrationCallbacks);
  
  if (g_UserGenerator.GetSkeletonCap().NeedPoseForCalibration()) {
    g_bNeedPose = TRUE;

    if (!g_UserGenerator.IsCapabilitySupported(XN_CAPABILITY_POSE_DETECTION)) {
      std::cerr << "Pose required, but not supported" 
		<< std::endl;
      throw 1;
    }

    g_UserGenerator.GetPoseDetectionCap().
      RegisterToPoseCallbacks(UserPose_PoseDetected, 
  			      NULL, NULL, 
  			      hPoseCallbacks);
    g_UserGenerator.GetSkeletonCap().GetCalibrationPose(g_strPose);
  }
  
  g_UserGenerator.GetSkeletonCap().SetSkeletonProfile(XN_SKEL_PROFILE_ALL);
  
  nRetVal = g_Context.StartGeneratingAll();
  CHECK_RC(nRetVal, "StartGenerating");
}

void Event::headTracker(){
  XnUserID users[1];
  XnUInt16 n_users = 16;
  XnSkeletonJointPosition p_head;

  g_UserGenerator.GetUsers(users,n_users);

  for(int i = 0 ; i < n_users ; i++){
    if(g_UserGenerator.GetSkeletonCap().IsTracking(users[i])){
      g_UserGenerator.GetSkeletonCap().GetSkeletonJointPosition(users[i],
								XN_SKEL_HEAD,
								p_head);
      _listener->headMoved(p_head.position.X,
      			   p_head.position.Y,
      			   p_head.position.Z);
      break;
    }
  }
}

void Event::setListener(HeadListener* l){
  _listener = l;
}

void Event::eventMainLoop(Drawer* drawer){
  SDL_Event event;
  bool quit = false;

  while(!quit){
    while(SDL_PollEvent(&event)){
      switch(event.type){
      case SDL_KEYDOWN:
	switch(event.key.keysym.sym){
	case SDLK_LEFT:
	  drawer->increaseX();
	  break;
	case SDLK_RIGHT:
	  drawer->decreaseX();
	  break;
	default:
	  break;
	}
	break;
      case SDL_QUIT:
	quit = true;
	break;
      default:
	break;
      }
    }
    update();
    drawer->drawScene();
    SDL_GL_SwapBuffers();
  }
}

void Event::update(){
  g_Context.WaitAndUpdateAll();
  headTracker();
}


Event::~Event(){

}
